from reinvent_scoring.scoring.score_components.pip.ratpk_pip import RatPKPiP
from reinvent_scoring.scoring.score_components.pip.pip_log_prediction_component import PiPLogPredictionComponent
from reinvent_scoring.scoring.score_components.pip.pip_prediction_component import PiPPredictionComponent
from reinvent_scoring.scoring.score_components.pip.qptuna_pip_model_component import QptunaPiPModelComponent
from reinvent_scoring.scoring.score_components.pip.string_pip_prediction_component import StringPiPPredictionComponent



from dataclasses import dataclass


@dataclass(frozen=True)
class ScoringFunctionComponentNameEnum:
    PARALLEL_ROCS_SIMILARITY = "parallel_rocs_similarity"
    SELECTIVITY = "selectivity"
    PREDICTIVE_PROPERTY = "predictive_property"
    ROCS_SIMILARITY = "rocs_similarity"
    MATCHING_SUBSTRUCTURE = "matching_substructure"
    TANIMOTO_SIMILARITY = "tanimoto_similarity"
    JACCARD_DISTANCE = "jaccard_distance"
    CUSTOM_ALERTS = "custom_alerts"
    QED_SCORE = "qed_score"
    MOLECULAR_WEIGHT = "molecular_weight"
    NUM_ROTATABLE_BONDS = "num_rotatable_bonds"
    NUM_HBD_LIPINSKI = "num_hbd_lipinski"
    NUM_HBA_LIPINSKI = "num_hba_lipinski"
    NUM_RINGS = "num_rings"
    NUM_AROMATIC_RINGS = "num_aromatic_rings"
    NUM_ALIPHATIC_RINGS = "num_aliphatic_rings"
    TPSA = "tpsa"
    SLOGP = "slogp"
    GRAPH_LENGTH = "graph_length"
    NUMBER_OF_STEREO_CENTERS = "number_of_stereo_centers"
    TOTAL_SCORE = "total_score" # there is no actual component corresponding to this type
    REACTION_FILTERS = "reaction_filters"

    # Link invent specific
    LINKER_EFFECTIVE_LENGTH = "linker_effective_length"
    LINKER_GRAPH_LENGTH = "linker_graph_length"
    LINKER_LENGTH_RATIO = "linker_length_ratio"
    LINKER_NUM_RINGS = "linker_num_rings"
    LINKER_NUM_ALIPHATIC_RINGS = "linker_num_aliphatic_rings"
    LINKER_NUM_AROMATIC_RINGS = "linker_num_aromatic_rings"
    LINKER_NUM_SP_ATOMS = "linker_num_sp_atoms"
    LINKER_NUM_SP2_ATOMS = "linker_num_sp2_atoms"
    LINKER_NUM_SP3_ATOMS = "linker_num_sp3_atoms"
    LINKER_NUM_HBA = "linker_num_hba"
    LINKER_NUM_HBD = "linker_num_hbd"
    LINKER_MOL_WEIGHT = "linker_mol_weight"
    LINKER_RATIO_ROTATABLE_BONDS = "linker_ratio_rotatable_bonds"

    #NOTE: components below are AZ specific
    SA_SCORE = "sa_score"
    AZDOCK = "azdock"
    DOCKSTREAM = "dockstream"
    ICOLOS = "icolos"
    AZ_LOGD74_PIP = "azlogd74"
    CACO2_INTR_PIP = "caco2-intrinsic-papp"
    CACO2_EFFLUX_PIP = "caco2-efflux"
    HH_CLINT_PIP = "hh-clint"
    HLM_CLINT_PIP = "hlm-clint"
    RH_CLINT_PIP = "rh-clint"
    SOLUBILITY_DD_PIP = "solubility-dd"
    HERG_PIP = "herg"
    KPUU_PIP = "rat-kpuu-brain"
    RAT_PK_PIP = "rat-pk"
    CLAB_TOP_20 = "clab_top_20"
    RA_SCORE = "rascore"
    AIZYNTH = "aizynth"
    QPTUNA_PIP_MODEL = "optuna-multi"
    THP1_CYTOTOXICITY = "thp1-class"



from typing import List

import numpy as np

from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.score_components.base_score_component import BaseScoreComponent
from reinvent_scoring.scoring.score_summary import ComponentSummary
from aizynthfinder.aizynthfinder import AiZynthExpander


class BuildingBlockAvailabilityComponent(BaseScoreComponent):
    """AiZynth one-step synthesis building block availability.

    Score is the ratio between
    the number of reactants in stock
    and the number of all reactants.

    If a molecule can be synthesized using different reactions,
    with different sets of reactants,
    the maximum ratio is used.

    This scoring component uses AiZynthFinder Expansion interface:
    https://molecularai.github.io/aizynthfinder/python_interface.html#expansion-interface
    """

    def __init__(self, parameters: ComponentParameters):
        super().__init__(parameters)

        configfile = self.parameters.specific_parameters[
            self.component_specific_parameters.AIZYNTH_CONFIG_FILE_PATH
        ]
        self._expander = self._set_up_expander(configfile)

    def calculate_score(self, molecules: List, step=-1) -> ComponentSummary:
        valid_smiles = self._chemistry.mols_to_smiles(molecules)
        score = self._score_smiles(valid_smiles)  # This is the main calculation.
        score_summary = ComponentSummary(total_score=score, parameters=self.parameters, raw_score=score)

        return score_summary

    def _score_one_smi(self, smi: str) -> float:
        stock = self._expander.config.stock

        reactions = self._expander.do_expansion(smi)

        ratios = []  # Collect all, in case there are alternative reactions.
        for reaction_tuple in reactions:
            precursors = reaction_tuple[0].reactants[0]
            if len(precursors) == 0:
                # Corner case - no reactants.
                # It implies that the template was not applicable on the query molecule,
                # or there was an error, and it was not possible to produce reactants.
                ratios.append(0)  # Assign the lowest possible score.
            else:
                in_stock = [mol in stock for mol in precursors]
                ratio_in_stock = sum(in_stock) / len(in_stock)
                ratios.append(ratio_in_stock)

        if len(ratios) > 0:
            max_ratio = max(ratios)  # Take the best.
        else:
            max_ratio = 0  # No building blocks, return the lowest possible score.

        return max_ratio

    def _score_smiles(self, smiles: List[str]) -> np.ndarray:
        results = [self._score_one_smi(smi) for smi in smiles]
        return np.array(results)

    def _set_up_expander(self, configfile: str) -> AiZynthExpander:
        expander = AiZynthExpander(configfile=configfile)
        expander.expansion_policy.select_first()
        expander.filter_policy.select_first()
        expander.config.stock.select_first()
        return expander



from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.score_components.link_invent.base_link_invent_component import BaseLinkInventComponent


class LinkerNumHBA(BaseLinkInventComponent):
    def __init__(self, parameters: ComponentParameters):
        super().__init__(parameters)

    def _calculate_linker_property(self, labeled_mol):
        return self._linker_descriptor.num_hba(labeled_mol)



from dataclasses import dataclass


@dataclass(frozen=True)
class ROCSSpecificParametersEnum():
    ROCS_INPUT = "rocs_input"
    INPUT_TYPE = "input_type"
    SHAPE_WEIGHT = "shape_weight"
    COLOR_WEIGHT = "color_weight"
    SIM_MEASURE = "similarity_measure"
    MAX_CPUS = "max_num_cpus"
    CUSTOM_CFF = "custom_cff"
    SAVE_ROCS_OVERLAYS = "save_rocs_overlays"
    ROCS_OVERLAYS_DIR = "rocs_overlays_dir"
    ROCS_OVERLAYS_PREFIX = "rocs_overlays_prefix"
    ENUM_STEREO = "enumerate_stereo"
    MAX_STEREO = "max_stereocenters"
    NEGATIVE_VOLUME = "negative_volume"
    PROTEIN_NEG_VOL_FILE = "protein_neg_vol_file"
    LIGAND_NEG_VOL_FILE = "ligand_neg_vol_file"
    MAX_CONFS = "max_confs"
    EWINDOW = "ewindow"



from dataclasses import dataclass


@dataclass(frozen=True)
class DiversityFilterEnum:
    IDENTICAL_TOPOLOGICAL_SCAFFOLD = "IdenticalTopologicalScaffold"
    IDENTICAL_MURCKO_SCAFFOLD = "IdenticalMurckoScaffold"
    SCAFFOLD_SIMILARITY = "ScaffoldSimilarity"
    NO_FILTER = "NoFilter"



from reinvent_scoring.scoring.score_components.link_invent.linker_effective_length import LinkerEffectiveLength
from reinvent_scoring.scoring.score_components.link_invent.linker_graph_length import LinkerGraphLength
from reinvent_scoring.scoring.score_components.link_invent.linker_length_ratio import LinkerLengthRatio
from reinvent_scoring.scoring.score_components.link_invent.linker_num_rings import LinkerNumRings
from reinvent_scoring.scoring.score_components.link_invent.linker_num_aliphatic_rings import LinkerNumAliphaticRings
from reinvent_scoring.scoring.score_components.link_invent.linker_num_aromatic_rings import LinkerNumAromaticRings
from reinvent_scoring.scoring.score_components.link_invent.linker_num_sp_atoms import LinkerNumSPAtoms
from reinvent_scoring.scoring.score_components.link_invent.linker_num_sp2_atoms import LinkerNumSP2Atoms
from reinvent_scoring.scoring.score_components.link_invent.linker_num_sp3_atoms import LinkerNumSP3Atoms
from reinvent_scoring.scoring.score_components.link_invent.linker_num_hba import LinkerNumHBA
from reinvent_scoring.scoring.score_components.link_invent.linker_num_hbd import LinkerNumHBD
from reinvent_scoring.scoring.score_components.link_invent.linker_mol_weight import LinkerMolWeight
from reinvent_scoring.scoring.score_components.link_invent.linker_ratio_rotatable_bonds import LinkerRatioRotatableBonds



import numpy as np
from typing import List
from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.score_components.pip.base_pip_model_component import BasePiPModelComponent


class StringPiPPredictionComponent(BasePiPModelComponent):
    """
    This class is to be used with pip models that return non-numeric (string mostly)
    values as predictions - casts raw value to float immediately.
    """

    def _parse_single_compound(self, compound):
        mapping = self.parameters.specific_parameters[self.component_specific_parameters.VALUE_MAPPING]
        value = compound["prediction"]

        return mapping.get(value, 0.0)



from reinvent_scoring.scoring.diversity_filters.lib_invent.identical_murcko_scaffold import IdenticalMurckoScaffold
from reinvent_scoring.scoring.diversity_filters.lib_invent.no_filter import NoFilter
from reinvent_scoring.scoring.diversity_filters.lib_invent.no_filter_with_penalty import NoFilterWithPenalty



from typing import List
from reinvent_scoring.scoring.component_parameters import ComponentParameters
from reinvent_scoring.scoring.score_components.pip.pip_prediction_component import PiPPredictionComponent


class QptunaPiPModelComponent(PiPPredictionComponent):

    def _format_data(self, smiles: List[str]) -> dict:
        molecules = [{"molData": smi, "id": f"{i}"} for i, smi in enumerate(smiles)]
        data = {
            "jsonData": {
                "data": molecules,
                "metadata": {
                    "molFormat":
                        "smiles"
                },
                "parameters": {
                    "artifact": self.parameters.specific_parameters.get(self.component_specific_parameters.ARTIFACT)
                }
            }
        }

        return data



from dataclasses import dataclass


@dataclass(frozen=True)
class TransformationTypeEnum:
    DOUBLE_SIGMOID = "double_sigmoid"
    SIGMOID = "sigmoid"
    REVERSE_SIGMOID = "reverse_sigmoid"
    RIGHT_STEP = "right_step"
    LEFT_STEP = "left_step"
    STEP = "step"
    CUSTOM_INTERPOLATION = "custom_interpolation"
    NO_TRANSFORMATION = "no_transformation"



from dataclasses import dataclass


@dataclass(frozen=True)
class ComponentSpecificParametersEnum:
    SCIKIT = "scikit"
    CLAB_INPUT_FILE = "clab_input_file"
    DESCRIPTOR_TYPE = "descriptor_type"
    TRANSFORMATION = "transformation"

    # structural components
    # ---------
    # AZDOCK
    AZDOCK_CONFPATH = "configuration_path"
    AZDOCK_DOCKERSCRIPTPATH = "docker_script_path"
    AZDOCK_ENVPATH = "environment_path"
    AZDOCK_DEBUG = "debug"

    # DockStream
    DOCKSTREAM_CONFPATH = "configuration_path"
    DOCKSTREAM_DOCKERSCRIPTPATH = "docker_script_path"
    DOCKSTREAM_ENVPATH = "environment_path"
    DOCKSTREAM_DEBUG = "debug"

    # ICOLOS
    ICOLOS_CONFPATH = "configuration_path"
    ICOLOS_EXECUTOR_PATH = "executor_path"
    ICOLOS_VALUES_KEY = "values_key"
    ICOLOS_DEBUG = "debug"
    #######################

    RAT_PK_PROPERTY = "rat_pk_property"
    CLAB_TOP_20_VALUE = "clab_top_20_value"
    ION_CLASS = "Ion class"
    CONTAINER_TYPE = "container_type"

    SMILES = "smiles"
    MODEL_PATH = "model_path"

    #######################
    ARTIFACT = "artifact"

    AIZYNTH_CONFIG_FILE_PATH = "aizynth_config_file_path"

    VALUE_MAPPING = "value_mapping"
